/*
 * Copyright (C) 2006-2007	Andre Beckedorf
 * 					 		<evilJazz _AT_ katastrophos _DOT_ net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <cstdio>
#include <cstdlib>
#include <unistd.h>

#include <qstring.h>
#include <qdatastream.h>
#include <qtextcodec.h>
#include <qregexp.h>

#ifdef WINDOWS
#include "wincompat.h"
#endif

#include "mediaidentifier.h"
#include "mediadatabase.h"
#include "mplayerbackend.h"
#include "configuration.h"
#include "debug.h"
#include "helpers.h"

#include <qbuffer.h>

#include <fileref.h>
#include <tstring.h>
#include <tmap.h>
#include <tstringlist.h>
#include <tbytevector.h>
#include <tag.h>
#include <mpegfile.h>
#include <attachedpictureframe.h>
#include <oggfile.h>
#include <vorbisfile.h>
#include <flacfile.h>
#include <oggflacfile.h>

#include <mp4file.h>
#include <mp4itunestag.h>

#ifndef QT4
#include <kmdcodec.h>
#endif

#include "compathack.h"

MediaIdentifier & MediaIdentifier::singleton()
{
    static MediaIdentifier instance;
    return instance;
}


MediaIdentifier::MediaIdentifier()
    : media_(NULL),
      fileScanningFinished_(false)
{
}

MediaIdentifier::~MediaIdentifier()
{
}

void MediaIdentifier::registerIdentifier(AbstractMediaIdentifier *identifier)
{
#ifdef QT4
    if (!identifiersMap_.contains(identifier->internalClassName()))
    {
        identifiers_.append(identifier);
        identifiersMap_.insert(identifier->internalClassName(), identifier);
    }
#else
    if (!identifiersMap_.find(identifier->internalClassName()))
    {
        identifiers_.append(identifier);
        identifiersMap_.insert(identifier->internalClassName(), identifier);
    }
#endif
}

void MediaIdentifier::identify(Media *media, const MediaLocation &location)
{
    bool tagsRead = false;
    bool audioPropertiesRead = false;

    QString extension = QFileInfo(location.toString()).suffix();

    media_ = media;
    fileScanningFinished_ = false;

    if (location.isLocalFile() && QFile::exists(location.toString()))
        loadTags(location, tagsRead, audioPropertiesRead);

    if (!tagsRead || !audioPropertiesRead || qConfig.isAmbiguousFormatExtension(extension))
    {
        for (int i = 0; i < identifiers_.count(); ++i)
        {
            if (identifiers_.at(i)->identify(media, location))
                break;
        }
    }

    media_ = NULL;
}

void MediaIdentifier::addMediadata(const QString &tag, const QString &value)
{
    media_->mediaMetadataExt()->addMediadata(tag, value);
}

bool MediaIdentifier::loadTags(const MediaLocation &location, bool &tagsWereReadable, bool &audioPropertiesWereReadable)
{
    TagLib::FileRef f(QFile::encodeName(location.toString()));

    tagsWereReadable = false;
    audioPropertiesWereReadable = false;

    if (!f.isNull() && f.tag())
    {
        TagLib::Tag *tag = f.tag();

        if (!tag->title().isNull())	addMediadata("TITLE", TStringToQString(tag->title()));
        if (!tag->album().isNull())	addMediadata("ALBUM", TStringToQString(tag->album()));
        if (!tag->artist().isNull())	addMediadata("ARTIST", TStringToQString(tag->artist()));
        if (!tag->genre().isNull())	addMediadata("GENRE", TStringToQString(tag->genre()));
        if (!tag->comment().isNull())	addMediadata("COMMENT", TStringToQString(tag->comment()));

        addMediadata("TRACK", QString::number(tag->track()));

        tagsWereReadable = true;
    }

    if (!f.isNull() && f.audioProperties())
    {
        TagLib::AudioProperties *properties = f.audioProperties();

        media_->mediaType_ = Media::MEDIATYPE_AUDIO;
        media_->length_ = properties->length();

        MediaAudio *audio = media_->mediaAudio();

        QString extension = QFileInfo(location.toString()).suffix().toLower();
        QString codec = "";

        // TODO: This lookup is by no means accurate. Improve!
        if (extension.startsWith("ogg"))
            codec = "Vorbis";
        else if (extension.startsWith("mp3"))
            codec = "MP3";
        else if (extension.startsWith("fla") || extension.startsWith("oga"))
            codec = "FLAC";
        else if (extension.startsWith("mpc"))
            codec = "Musepack";
        else if (extension.startsWith("aac"))
            codec = "AAC";
        else if (extension.startsWith("mp4") || extension.startsWith("m4"))
            codec = "MP4";
        else if (extension.startsWith("wma") || extension.startsWith("asf"))
            codec = "Windows Media Audio";
        else if (extension.startsWith("wv"))
            codec = "WavPack";
        else if (extension.startsWith("wav"))
            codec = "WAVE";

        audio->codec = codec;
        audio->bitrate = properties->bitrate() * 1000;
        audio->sampleRate = properties->sampleRate();
        audio->channels = properties->channels();

        audioPropertiesWereReadable = true;
    }

    return false;
}

void MediaIdentifier::dumpXiphComment(TagLib::Ogg::XiphComment *xiphComment)
{
    DGUARDMETHOD;

    TagLib::Ogg::FieldListMap::ConstIterator it = xiphComment->fieldListMap().begin();

    for (; it != xiphComment->fieldListMap().end(); ++it)
    {
        TagLib::String fieldName = (*it).first;
        TagLib::StringList values = (*it).second;

        TagLib::StringList::ConstIterator valuesIt = values.begin();

        for (; valuesIt != values.end(); ++valuesIt)
            DPRINTF("%s = %s", fieldName.toCString(), (*valuesIt).toCString());
    }
}

bool MediaIdentifier::loadCoverArtFromXiphComment(TagLib::Ogg::XiphComment *xiphComment, QImage &coverArtImage)
{
    DGUARDMETHODTIMED;
    bool result = false;

    if (xiphComment->contains("COVERART"))
    {
        DPRINTF("xiphComment contains COVERART.");

        TagLib::ByteVector coverArtBase64Data = xiphComment->fieldListMap()["COVERART"].front().data(TagLib::String::Latin1);

        // create a shallow copy QByteArray wrapper...
#ifdef QT4
        QByteArray coverArtBase64DataInput(QByteArray::fromRawData(coverArtBase64Data.data(), coverArtBase64Data.size()));
        QByteArray coverArtRawDataOutput(QByteArray::fromBase64(coverArtBase64DataInput));
#else
        QByteArray coverArtBase64DataInput;
        coverArtBase64DataInput.setRawData(coverArtBase64Data.data(), coverArtBase64Data.size());

        QByteArray coverArtRawDataOutput;
        KCodecs::base64Decode(coverArtBase64DataInput, coverArtRawDataOutput);

        // the following is required for resetting the byte array, as stated in the Qt2/Qt3 docs...
        coverArtBase64DataInput.resetRawData(coverArtBase64Data.data(), coverArtBase64Data.size());
#endif

        result = coverArtImage.loadFromData(coverArtRawDataOutput);
        DPRINTF("%s", result ? "Read image successfully." : "Could not decode image.");
    }

    return result;
}

bool MediaIdentifier::loadCoverArtFromID3v2(TagLib::ID3v2::Tag *id3v2Tag, QImage &coverArtImage)
{
    DGUARDMETHODTIMED;
    bool result = false;

    TagLib::ID3v2::FrameList frameList = id3v2Tag->frameList("APIC");

    if (!frameList.isEmpty())
    {
        TagLib::ID3v2::AttachedPictureFrame *attachedPictureFrame = static_cast<TagLib::ID3v2::AttachedPictureFrame *>(frameList.front());

        if (attachedPictureFrame->picture().size() > 0)
        {
            DPRINTF("ID3v2 header contains APIC frame.");
            result = coverArtImage.loadFromData((const uchar *)attachedPictureFrame->picture().data(), attachedPictureFrame->picture().size());
            DPRINTF("%s", result ? "Read image successfully." : "Could not decode image.");
        }
    }

    return result;
}

bool MediaIdentifier::loadCoverArtFromFile(const QString &filename, QImage &coverArtImage)
{
    DGUARDMETHODFTIMED("filename: %s", (const char *)filename.toUtf8());

    bool result = false;

    //if (!QFile(filename).exists())
    //	return false;

    QString extension = QFileInfo(filename).suffix().toLower();

    // Check for a cover art image in ID3v2 APIC frame
    if (extension.indexOf("mp3") != -1)
    {
        TagLib::MPEG::File f(QFile::encodeName(filename));

        if (f.isValid() && f.ID3v2Tag())
            result = loadCoverArtFromID3v2(f.ID3v2Tag(), coverArtImage);
    }
    // Check for a cover art image in Ogg Container in XiphComment
    else if ((extension.indexOf("ogg") != -1) ||
             (extension.indexOf("oga") != -1))
    {
        TagLib::Ogg::Vorbis::File f(QFile::encodeName(filename));

        result = f.isValid() && f.tag() && loadCoverArtFromXiphComment(f.tag(), coverArtImage);

        DHOP(dumpXiphComment(f.tag()));

        if (!result)
        {
            TagLib::Ogg::FLAC::File f(QFile::encodeName(filename));

            if (f.isValid() && f.tag())
                result = loadCoverArtFromXiphComment(f.tag(), coverArtImage);
        }
    }
    // Check for a cover art image either via ID3v2 APIC frame or in XiphComment
    else if (extension.indexOf("flac") != -1)
    {
        TagLib::FLAC::File f(QFile::encodeName(filename));

        if (f.isValid())
        {
            result = f.ID3v2Tag() && loadCoverArtFromID3v2(f.ID3v2Tag(), coverArtImage);

            if (!result)
                result = f.xiphComment() && loadCoverArtFromXiphComment(f.xiphComment(), coverArtImage);
        }
    }
    // Check for a M4A iTunes cover art image
    else if ((extension.indexOf("m4a") != -1) ||
            (extension.indexOf("aac") != -1) ||
            (extension.indexOf("m4b") != -1) ||
            (extension.indexOf("m4p") != -1) ||
            (extension.indexOf("mp4") != -1) ||
            (extension.indexOf("m4v") != -1) ||
            (extension.indexOf("mp4v") != -1))
    {
        TagLib::MP4::File f(QFile::encodeName(filename));

        if (f.isValid() && f.tag())
        {
            TagLib::MP4::Tag *tag = static_cast<TagLib::MP4::Tag *>(f.tag());

            if (!tag->cover().isNull() && !tag->cover().isEmpty())
            {
                DPRINTF("M4A file contains iTunes covr atom.");
                result = coverArtImage.loadFromData((const uchar *)tag->cover().data(), tag->cover().size());
                DPRINTF("%s", result ? "Read image successfully." : "Could not decode image.");
            }
        }
    }

    return result;
}
